import numpy as np
from .pted import pted, pted_coverage_test


def test():
    np.random.seed(42)
    # example 2 sample test
    D = 300
    for _ in range(20):
        x = np.random.normal(size=(100, D))
        y = np.random.normal(size=(100, D))
        p = pted(x, y)
        assert p > 1e-4 and p < 0.9999, f"p-value {p} is not in the expected range (U(0,1))"

    x = np.random.normal(size=(100, D))
    y = np.random.uniform(size=(100, D))
    p = pted(x, y)
    assert p < 1e-4, f"p-value {p} is not in the expected range (~0)"

    x = np.random.normal(size=(100, D))
    p = pted(x, x)
    assert p > 0.9999, f"p-value {p} is not in the expected range (~1)"

    # example coverage
    n_sims = 100
    n_samples = 100
    D = 2
    g = []
    s_corr = []
    s_over = []
    s_under = []
    for _ in range(n_sims):
        loc = np.random.normal(size=(D)) * 10
        scale = np.random.uniform(size=(D)) * 10 + 1
        g.append(np.random.normal(loc=loc, scale=scale, size=(D)))
        s_corr.append(np.random.normal(loc=loc, scale=scale, size=(n_samples, D)))
        s_over.append(np.random.normal(loc=loc, scale=scale / 2, size=(n_samples, D)))
        s_under.append(np.random.normal(loc=loc, scale=scale * 2, size=(n_samples, D)))
    g = np.array(g)
    s_corr = np.stack(s_corr, axis=1)
    s_over = np.stack(s_over, axis=1)
    s_under = np.stack(s_under, axis=1)

    # correct
    p = pted_coverage_test(g, s_corr, permutations=200)
    assert p > 1e-4 and p < 0.9999, f"p-value {p} is not in the expected range (U(0,1))"
    # overconfident
    p = pted_coverage_test(g, s_over, permutations=200)
    assert p < 1e-4, f"p-value {p} is not in the expected range (~0)"
    # underconfident
    p = pted_coverage_test(g, s_under, permutations=200)
    assert p > 0.9999, f"p-value {p} is not in the expected range (~1)"

    print("Tests passed!")



import numpy as np
from scipy.spatial.distance import cdist
import torch

__all__ = ["_pted_numpy", "_pted_chunk_numpy", "_pted_torch", "_pted_chunk_torch"]


def _energy_distance_precompute(D, nx, ny):
    Exx = D[:nx, :nx].sum() / nx**2
    Eyy = D[nx:, nx:].sum() / ny**2
    Exy = D[:nx, nx:].sum() / (nx * ny)
    return 2 * Exy - Exx - Eyy


def _energy_distance_estimate(x, y, chunk_size, chunk_iter, metric="euclidean"):
    is_torch = isinstance(x, torch.Tensor)

    E_est = []
    for _ in range(chunk_iter):
        # Randomly sample a chunk of data
        idx = np.random.choice(len(x), size=min(len(x), chunk_size), replace=False)
        if is_torch:
            idx = torch.tensor(idx, device=x.device)
        x_chunk = x[idx]
        idy = np.random.choice(len(y), size=min(len(y), chunk_size), replace=False)
        if is_torch:
            idy = torch.tensor(idy, device=y.device)
        y_chunk = y[idy]

        # Compute the distance matrix
        if is_torch:
            z_chunk = torch.cat((x_chunk, y_chunk), dim=0)
        else:
            z_chunk = np.concatenate((x_chunk, y_chunk), axis=0)
        dmatrix = cdist(z_chunk, z_chunk, metric=metric)

        # Compute the energy distance
        E_est.append(_energy_distance_precompute(dmatrix, len(x_chunk), len(y_chunk)))
        if is_torch:
            E_est[-1] = E_est[-1].item()
    return np.mean(E_est)


def _pted_chunk_numpy(x, y, permutations=100, metric="euclidean", chunk_size=100, chunk_iter=10):
    assert np.all(np.isfinite(x)) and np.all(np.isfinite(y)), "Input contains NaN or Inf!"
    nx = len(x)

    test_stat = _energy_distance_estimate(x, y, chunk_size, chunk_iter, metric=metric)
    permute_stats = []
    for _ in range(permutations):
        z = np.concatenate((x, y), axis=0)
        z = z[np.random.permutation(len(z))]
        x, y = z[:nx], z[nx:]
        permute_stats.append(_energy_distance_estimate(x, y, chunk_size, chunk_iter, metric=metric))
    return test_stat, permute_stats


def _pted_chunk_torch(x, y, permutations=100, metric="euclidean", chunk_size=100, chunk_iter=10):
    assert torch.all(torch.isfinite(x)) and torch.all(
        torch.isfinite(y)
    ), "Input contains NaN or Inf!"
    nx = len(x)

    test_stat = _energy_distance_estimate(x, y, chunk_size, chunk_iter, metric=metric)
    permute_stats = []
    for _ in range(permutations):
        z = torch.cat((x, y), dim=0)
        z = z[torch.randperm(len(z))]
        x, y = z[:nx], z[nx:]
        permute_stats.append(_energy_distance_estimate(x, y, chunk_size, chunk_iter, metric=metric))
    return test_stat, permute_stats


def _pted_numpy(x, y, permutations=100, metric="euclidean"):
    z = np.concatenate((x, y), axis=0)
    assert np.all(np.isfinite(z)), "Input contains NaN or Inf!"
    dmatrix = cdist(z, z, metric=metric)
    assert np.all(
        np.isfinite(dmatrix)
    ), "Distance matrix contains NaN or Inf! Consider using a different metric or normalizing values to be more stable (i.e. z-score norm)."
    nx = len(x)
    ny = len(y)

    test_stat = _energy_distance_precompute(dmatrix, nx, ny)
    permute_stats = []
    for _ in range(permutations):
        I = np.random.permutation(len(z))
        dmatrix = dmatrix[I][:, I]
        permute_stats.append(_energy_distance_precompute(dmatrix, nx, ny))
    return test_stat, permute_stats


@torch.no_grad()
def _pted_torch(x, y, permutations=100, metric="euclidean"):
    z = torch.cat((x, y), dim=0)
    assert torch.all(torch.isfinite(z)), "Input contains NaN or Inf!"
    if metric == "euclidean":
        metric = 2.0
    dmatrix = torch.cdist(z, z, p=metric)
    assert torch.all(
        torch.isfinite(dmatrix)
    ), "Distance matrix contains NaN or Inf! Consider using a different metric or normalizing values to be more stable (i.e. z-score norm)."
    nx = len(x)
    ny = len(y)

    test_stat = _energy_distance_precompute(dmatrix, nx, ny).item()
    permute_stats = []
    for _ in range(permutations):
        I = torch.randperm(len(z))
        dmatrix = dmatrix[I][:, I]
        permute_stats.append(_energy_distance_precompute(dmatrix, nx, ny).item())
    return test_stat, permute_stats



from typing import Union, Optional
import numpy as np
from scipy.stats import chi2 as chi2_dist
from torch import Tensor

from .utils import _pted_torch, _pted_numpy, _pted_chunk_torch, _pted_chunk_numpy

__all__ = ["pted", "pted_coverage_test"]


def pted(
    x: Union[np.ndarray, Tensor],
    y: Union[np.ndarray, Tensor],
    permutations: int = 1000,
    metric: str = "euclidean",
    return_all: bool = False,
    chunk_size: Optional[int] = None,
    chunk_iter: Optional[int] = None,
):
    """
    Two sample null hypothesis test using a permutation test on the energy
    distance.

    A "two sample test" is a statistical test that compares two samples to
    determine if they come from the same distribution. The null hypothesis is
    that the two samples come from the same distribution. A permutation test is
    a non-parametric test that compares a test statistic (in this case, the
    energy distance) to that same statistic computed on random re-shuffling
    (permutations) of the data. Under the null hypothesis, x and y were drawn
    from the same distribution so test statistic should be randomly distributed
    among the permutation statistics. If the test statistic is significantly
    larger than the permuted statistics, the p-value will be very small. Before
    running pted, you should choose a threshold at which you will reject the
    null. for example, if you choose a threshold of 0.01, you will reject the
    null hypothesis if the p-value is less than 0.01. However, note that in this
    case you will reject the null 1% of the time even if the null is true. This
    is a trade-off between false positives and false negatives.

    Here is a pseudo-code description of the algorithm:
        test_stat = energy_distance(x, y)
        permute_stats = []
        for i in range(permutations):
            z = concatenate(x, y)
            z = shuffle(z)
            x, y = z[:nx], z[nx:]
            permute_stats.append(energy_distance(x, y))
        p = mean(permute_stats > test_stat)
        return p

    Example
    -------
        import numpy as np
        from pted import pted

        # Generate two samples from the same distribution
        x = np.random.normal(size=(100, 10))
        y = np.random.normal(size=(100, 10))

        p = pted(x, y)

        print(f"p-value: {p}") # expect p in U(0,1)

    Parameters
    ----------
        x (Union[np.ndarray, Tensor]): first set of samples. Shape (N, *D)
        y (Union[np.ndarray, Tensor]): second set of samples. Shape (M, *D)
        permutations (int): number of permutations to run. This determines how
            accurately the p-value is computed.
        metric (str): distance metric to use. See scipy.spatial.distance.cdist
            for the list of available metrics with numpy. See torch.cdist when
            using PyTorch, note that the metric is passed as the "p" for
            torch.cdist and therefore is a float from 0 to inf.
        return_all (bool): if True, return the test statistic and the permuted
            statistics. If False, just return the p-value. bool (False by default)
        chunk_size (Optional[int]): if not None, use chunked energy distance
            estimation. This is useful for large datasets. The chunk size is the
            number of samples to use for each chunk. If None, use the full
            dataset.
        chunk_iter (Optional[int]): The chunk iter is the number of iterations
            to use with the given chunk size.

    Note
    ----
        PTED has O(n^2 * D * P) time complexity, where n is the number of
        samples in x and y, D is the number of dimensions, and P is the number
        of permutations. For large datasets this can get unwieldy, so chunking
        is recommended. For chunking, the energy distance will be estimated at
        each iteration rather than fully computed. To estimate the energy
        distance, we take `chunk_size` sub-samples from x and y, and compute the
        energy distance on those sub-samples. This is repeated `chunk_iter`
        times, and the average is taken. This is a trade-off between speed and
        accuracy. The larger the chunk size and larger chunk_iter, the more
        accurate the estimate, but the slower the computation. PTED remains an
        exact p-value test even when chunking, it simply becomes less sensitive
        to the difference between x and y. The chunked pted test has time
        complexity O(c^2 * I * D * P), where c is the chunk size, I is the
        number of iterations, D is the number of dimensions, and P is the number
        of permutations. For chunking to be worth it you should have c^2 * I << n^2.
    """
    assert type(x) == type(y), f"x and y must be of the same type, not {type(x)} and {type(y)}"
    assert len(x.shape) >= 2, f"x must be at least 2D, not {x.shape}"
    assert len(y.shape) >= 2, f"y must be at least 2D, not {y.shape}"
    assert (
        x.shape[1:] == y.shape[1:]
    ), f"x and y samples must have the same shape (past first dim), not {x.shape} and {y.shape}"
    if len(x.shape) > 2:
        x = x.reshape(x.shape[0], -1)
    if len(y.shape) > 2:
        y = y.reshape(y.shape[0], -1)

    if isinstance(x, Tensor) and chunk_size is not None:
        test, permute = _pted_chunk_torch(
            x,
            y,
            permutations=permutations,
            metric=metric,
            chunk_size=chunk_size,
            chunk_iter=chunk_iter,
        )
    elif isinstance(x, Tensor):
        test, permute = _pted_torch(x, y, permutations=permutations, metric=metric)
    elif chunk_size is not None:
        test, permute = _pted_chunk_numpy(
            x,
            y,
            permutations=permutations,
            metric=metric,
            chunk_size=chunk_size,
            chunk_iter=chunk_iter,
        )
    else:
        test, permute = _pted_numpy(x, y, permutations=permutations, metric=metric)

    if return_all:
        return test, permute

    # Compute p-value
    return np.mean(np.array(permute) > test)


def pted_coverage_test(
    g: Union[np.ndarray, Tensor],
    s: Union[np.ndarray, Tensor],
    permutations: int = 1000,
    metric: str = "euclidean",
    return_all: bool = False,
    chunk_size: Optional[int] = None,
    chunk_iter: Optional[int] = None,
):
    """
    Coverage test using a permutation test on the energy distance.

    A "coverage test" is a statistical test that determines if the posterior
    samples (s) cover the ground truth samples (g) with the correct uncertainty.
    By "correct uncertainty" we mean that any region R which contains a% of the
    posterior samples should contain the ground truth g with a% probability. A
    posterior s which is "overconfident" will have very little variability in
    it's samples, therefore the ground truth g will tend to be far off from the
    samples in relative terms. This will lead to a low p-value. A posterior s
    which is "underconfident" will have too much variability in it's samples,
    therefore the ground truth g will be enclosed too well within the
    distribution of samples. This will lead to a high p-value. The null
    hypothesis is that the posterior samples cover the ground truth samples with
    the correct uncertainty. If the null hypothesis is true, the p-value will be
    distributed as U(0,1).

    To perform this test, we compute the pted p-value for each simulation
    independently. We then compute the p-value under the null hypothesis that
    the pted p-values are distributed as U(0,1). This is done by computing the
    chi-squared statistic of the p-values (which for U(0,1) means chi2 = -2 *
    log(p)). The total p-value is then computed as 1 - chi2_cdf(sum(chi2), 2 *
    n_sims). Note, that because p is computed with a finite number of
    iterations, it is possible that p=0 in which case log(p) = -inf. To handle
    this, we set p=1/n_permutations. This is essentially the smallest p-value
    estimate reasonable for n_permutations.


    Example Usage
    ----------------
        import numpy as np
        from pted import pted_coverage_test

        # Generate mock ground truth samples (n_simulations, n_dimensions)
        g = np.random.normal(size=(100, 10))

        # Generate mock posterior samples (n_samples, n_simulations, n_dimensions)
        s = np.random.normal(size=(200, 100, 10))

        p = pted_coverage_test(g, s)

        print(f"p-value: {p}") # expect p in U(0,1)


    Parameters
    ----------
        g (Union[np.ndarray, Tensor]): Ground truth samples. Shape (n_sims, *D)
        s (Union[np.ndarray, Tensor]): Posterior samples. Shape (n_samples, n_sims, *D)
        permutations (int): number of permutations to run. This determines how
            accurately the p-value is computed.
        metric (str): distance metric to use. See scipy.spatial.distance.cdist
            for the list of available metrics with numpy. See torch.cdist when using
            PyTorch, note that the metric is passed as the "p" for torch.cdist and
            therefore is a float from 0 to inf.
        return_all (bool): if True, return the test statistic and the permuted
            statistics. If False, just return the p-value. bool (False by
            default)
        chunk_size (Optional[int]): if not None, use chunked energy distance
            estimation. This is useful for large datasets. The chunk size is the
            number of samples to use for each chunk. If None, use the full
            dataset.
        chunk_iter (Optional[int]): The chunk iter is the number of iterations
            to use with the given chunk size.

    Note
    ----
        PTED has O(n^2 * D * P) time complexity, where n is the number of
        samples in x and y, D is the number of dimensions, and P is the number
        of permutations. For large datasets this can get unwieldy, so chunking
        is recommended. For chunking, the energy distance will be estimated at
        each iteration rather than fully computed. To estimate the energy
        distance, we take `chunk_size` sub-samples from x and y, and compute the
        energy distance on those sub-samples. This is repeated `chunk_iter`
        times, and the average is taken. This is a trade-off between speed and
        accuracy. The larger the chunk size and larger chunk_iter, the more
        accurate the estimate, but the slower the computation. PTED remains an
        exact p-value test even when chunking, it simply becomes less sensitive
        to the difference between x and y. The chunked pted test has time
        complexity O(c^2 * I * D * P), where c is the chunk size, I is the
        number of iterations, D is the number of dimensions, and P is the number
        of permutations. For chunking to be worth it you should have c^2 * I << n^2.
    """
    nsamp, nsim, *D = s.shape
    assert (
        g.shape == s.shape[1:]
    ), f"g and s must have the same shape (past first dim of s), not {g.shape} and {s.shape}"
    if len(s.shape) > 3:
        s = s.reshape(nsamp, nsim, -1)
    g = g.reshape(1, nsim, -1)

    test_stats = []
    permute_stats = []
    for i in range(nsim):
        test, permute = pted(
            g[:, i],
            s[:, i],
            permutations=permutations,
            metric=metric,
            return_all=True,
            chunk_size=chunk_size,
            chunk_iter=chunk_iter,
        )
        test_stats.append(test)
        permute_stats.append(permute)
    test_stats = np.array(test_stats)
    permute_stats = np.array(permute_stats)

    if return_all:
        return test_stats, permute_stats

    # Compute p-values
    pvals = np.mean(permute_stats > test_stats[:, None], axis=1)
    pvals[pvals == 0] = 1.0 / permutations  # handle pvals == 0
    chi2 = -2 * np.log(pvals)
    return 1 - chi2_dist.cdf(np.sum(chi2), 2 * nsim)



import pted
import torch
import numpy as np


def test_inputs_extra_dims():
    np.random.seed(42)
    # Test with numpy arrays
    x = np.random.normal(size=(100, 30, 30))
    y = np.random.normal(size=(100, 30, 30))
    p = pted.pted(x, y)
    assert p > 1e-4 and p < 0.9999, f"p-value {p} is not in the expected range (U(0,1))"

    # Test with torch tensors
    g = torch.randn(100, 30, 30)
    s = torch.randn(50, 100, 30, 30)
    p = pted.pted_coverage_test(g, s)
    assert p > 1e-4 and p < 0.9999, f"p-value {p} is not in the expected range (U(0,1))"


def test_pted_main():
    pted.test()


def test_pted_torch():

    # Set the random seed for reproducibility
    torch.manual_seed(42)
    np.random.seed(42)

    # example 2 sample test
    D = 300
    for _ in range(20):
        x = torch.randn(100, D)
        y = torch.randn(100, D)
        p = pted.pted(x, y)
        assert p > 1e-4 and p < 0.9999, f"p-value {p} is not in the expected range (U(0,1))"

    x = torch.randn(100, D)
    y = torch.rand(100, D)
    p = pted.pted(x, y)
    assert p < 1e-4, f"p-value {p} is not in the expected range (~0)"

    x = torch.randn(100, D)
    t, p = pted.pted(x, x, return_all=True)
    p = np.mean(np.array(p) > t)
    assert p > 0.9999, f"p-value {p} is not in the expected range (~1)"


def test_pted_coverage_full():
    g = np.random.normal(size=(100, 10))  # ground truth (n_simulations, n_dimensions)
    s = np.random.normal(
        size=(200, 100, 10)
    )  # posterior samples (n_samples, n_simulations, n_dimensions)

    test, permute = pted.pted_coverage_test(g, s, permutations=100, return_all=True)
    assert test.shape == (100,)
    assert permute.shape == (100, 100)


def test_pted_chunk_torch():
    np.random.seed(42)
    torch.manual_seed(42)

    # example 2 sample test
    D = 10
    x = torch.randn(1000, D)
    y = torch.randn(1000, D)
    p = pted.pted(x, y, chunk_size=100, chunk_iter=10)
    assert p > 1e-4 and p < 0.9999, f"p-value {p} is not in the expected range (U(0,1))"

    y = torch.rand(1000, D)
    p = pted.pted(x, y, chunk_size=100, chunk_iter=10)
    assert p < 1e-4, f"p-value {p} is not in the expected range (~0)"


def test_pted_chunk_numpy():
    np.random.seed(42)

    # example 2 sample test
    D = 10
    x = np.random.normal(size=(1000, D))
    y = np.random.normal(size=(1000, D))
    p = pted.pted(x, y, chunk_size=100, chunk_iter=10)
    assert p > 1e-4 and p < 0.9999, f"p-value {p} is not in the expected range (U(0,1))"

    y = np.random.uniform(size=(1000, D))
    p = pted.pted(x, y, chunk_size=100, chunk_iter=10)
    assert p < 1e-4, f"p-value {p} is not in the expected range (~0)"



from .pted import pted, pted_coverage_test
from .tests import test
from ._version import version as __version__  # noqa

__author__ = "Connor Stone"
__email__ = "connorstone628@gmail.com"

__all__ = [
    "pted",
    "pted_coverage_test",
    "test",
    "__version__",
    "__author__",
    "__email__",
]
